/* File:  memory.c
 *
 *  Copyright:  2005, Tres Melton (until I submit it to mej@Eterm)
 *
 */

#define _XOPEN_SOURCE 600
#define _GNU_SOURCE
#define __USE_XOPEN2K
#include <stdlib.h>  

#include <stdio.h>  
#include <malloc.h>
#include "memory.h"

#if !defined DEBUG_ALIGNED 
#define DEBUG_ALIGNED 0
#endif

//  A macro from somewhere??
//#define IS_ALIGNED_128(val) (!((val) & 0x15))

#define forever for(;;)

aligned_mem_list amem_list;

void error_exit( const char *str )
{
  printf( "Error: %s\n", str );
  exit( 1 );
}

int my_alignof( void* ptr )
{
  int alignment = 1;
  unsigned long temp_ptr = ( long ) ptr;
  
//  printf( "temp_ptr: %lx\n", temp_ptr );
  while(!( temp_ptr & 0x1 ))
  {
//    printf( "temp_ptr: %lx\n", temp_ptr );
    alignment <<= 1;
    temp_ptr >>= 1;
  }
  return( alignment );
}

int my_req_alignof( void *ptr )
{
  aligned_memory *align_ptr = amem_list.head;
  int element = 0;
  
  while( align_ptr )
  {
    element++;
//    printf( "aptr:        0x%012lX \n", (long) align_ptr );
//    printf( "aptr->:      0x%012lX \n", (long) align_ptr -> align_ptr );
//    printf( "ptr:         0x%012lX \n", (long) ptr );
    if( align_ptr -> align_ptr == ptr )
    {
      return( align_ptr -> alignment );
    }
    align_ptr = align_ptr -> next;
  }
//  printf( "ptr:         0x%012lX (not found)\n", (long) ptr );
  error_exit( "No memory region found to query alignment of." );
  return 0;	/*  Otherwise it complains about control reaching the end of a non-void function.  */
}

void* malloc_aligned( const size_t size, const size_t alignment )
{
  void *data;
  aligned_memory *ptr;
  unsigned long temp = 1;
  
  //  sanity checks ( size, alignment )
  //    ensure that alignment = 2^x  (only 1 bit set)
  forever
  {
    if( temp == alignment )
      break;
    temp *= 2;
    if( temp > alignment )
      return( NULL );
      //  alignment != 2^x where x is whole and non-negative
  }

  //  This should be able to align everything that gets handed to it.
  data = malloc( size + alignment - 1 );
  if( !data )
    return( NULL );
  ptr = ( aligned_memory* ) malloc( sizeof( aligned_memory ));
  if( !ptr )
  {
    free( data );
    return( NULL );
  }
  ptr -> orig_ptr = data;
  ptr -> alignment = alignment;
//  printf( "Orig:        0x%012lX (Alignment: %d-%d)\n", (long) ptr -> orig_ptr, (int) __alignof__( ptr -> orig_ptr ), my_alignof( ptr -> orig_ptr ));


//  temp = ((((long) data ) + alignment - 1 ) & (~(( long ) ( alignment - 1 ))));

//  printf( "data-ptr:    0x%012lX \n", (long) data );
  temp = alignment - 1;
//  printf( "temp:        0x%012lX \n", (long) temp );
  temp += ( long ) data;
//  printf( "temp:        0x%012lX \n", (long) temp );
  temp &= (~(( long ) ( alignment - 1 )));
//  printf( "temp:        0x%012lX \n", (long) temp );

  ptr -> align_ptr = ( aligned_memory* ) temp;

//  ptr -> align_ptr = (( data + alignment - 1 ) & (( long ) ~( alignment - 1 )));


  ptr -> next = NULL;
  // insert ptr into the list
//  if( !( amem_list.head & amem_list.tail ))
////  printf( "Aligned:     0x%012lX (Alignment: %d-%d)\n", (long) ptr -> align_ptr, (int) __alignof__( ptr -> align_ptr ), my_alignof( ptr -> align_ptr ));
  if(( !amem_list.head && !amem_list.tail ))
  {
    // First element
    amem_list.number = 1;
    amem_list.head = ptr;
    ptr -> prev = NULL;
  }
  else
  {
    ptr -> prev = amem_list.tail;
    ptr -> prev -> next = ptr;
  }
  ptr -> next = NULL;  
  amem_list.tail = ptr;
  amem_list.number++;

  return( ptr -> align_ptr );
}
 
void free_aligned( void *ptr )
{
  aligned_memory *align_ptr = amem_list.head;
  int element = 0;
  
  while( align_ptr )
  {
    element++;
//    printf( "aptr:        0x%012lX \n", (long) align_ptr );
//    printf( "aptr->:      0x%012lX \n", (long) align_ptr -> align_ptr );
//    printf( "ptr:         0x%012lX \n", (long) ptr );
    if( align_ptr -> align_ptr == ptr )
    {
//      printf( "removing:    0x%012lX (%d)\n", (long) ptr, element );
      if( align_ptr -> prev )
        align_ptr -> prev -> next = align_ptr -> next;
      if( align_ptr -> next )
        align_ptr -> next -> prev = align_ptr -> prev;
      if( amem_list.head == align_ptr )
        amem_list.head = align_ptr -> next;
      if( amem_list.tail == align_ptr )
        amem_list.tail = align_ptr -> prev;
      free( align_ptr -> orig_ptr );
      free( align_ptr );
      amem_list.number--;
      return;
    }
    align_ptr = align_ptr -> next;
  }
//  printf( "ptr:         0x%012lX (not found)\n", (long) ptr );
  error_exit( "No memory region found to free." );
}

void init_aligned( void )
{
//  if(( amem_list.number == 0 ) && ( amem_list.head == NULL ) && ( amem_list.tail == NULL ))
//    error_exit( "Probably Already initialized" );
  amem_list.number = 0;
  amem_list.head = NULL;
  amem_list.tail = NULL;
}

void clean_aligned( void )
{
  aligned_memory *ptr;
  
  ptr = amem_list.head;
  if( ptr )
    printf( "List not empty.\n" );	// Not important if intended.
  while( ptr ) 
  {
    free( ptr -> orig_ptr );
    if( ptr -> next == NULL )
    {
      free( ptr );
      break;
    }
    ptr = ptr -> next;
    free( ptr -> prev );
  }
  amem_list.number = 0;
  amem_list.head = NULL;
  amem_list.tail = NULL;
}
 
#if defined DEBUG_ALIGNED && DEBUG_ALIGNED == 1

//#define COUNT		100
//#define COUNT		1000000
//time on an amd64 3500+
//real    0m5.892s
//user    0m2.619s
//sys     0m2.964s
#define COUNT         10000000
//#define COUNT		1000000000

#define ELEMENTS	1000
#define MAX_SIZE	1000000

int main( void )
{
  long   *data1;
  int    *data2;
  double *data3;
  char   *data4;
  void   *data5;
  int     index;
  int     count = 0;
  void   *array[ ELEMENTS ];

  init_aligned();
  printf( "-----------------------------------------------------------\n" );
  data1 = malloc_aligned(( 50 * sizeof( long )), 16 );
  if( !data1 )
    error_exit( "An error occured while allocating Data1." );
  printf( "Data1:       0x%012lX   __alignof__() = %-3d   my_alignof() = %3d\n", (long) data1, (int) __alignof__( data1 ), my_alignof( data1 ));
  data2 = malloc_aligned(( 50 * sizeof( int )), 16 );
  if( !data2 )
    error_exit( "An error occured while allocating Data2." );
  printf( "Data2:       0x%012lX   __alignof__() = %-3d   my_alignof() = %3d\n", (long) data2, (int) __alignof__( data2 ), my_alignof( data2 ));
  data3 = malloc_aligned(( 50 * sizeof( double )), 16 );
  if( !data3 )
    error_exit( "An error occured while allocating Data3." );
  printf( "Data3:       0x%012lX   __alignof__() = %-3d   my_alignof() = %3d\n", (long) data3, (int) __alignof__( data3 ), my_alignof( data3 ));
  data4 = malloc_aligned(( 50 * sizeof( char )), 16 );
  if( !data4 )
    error_exit( "An error occured while allocating Data4." );
  printf( "Data4:       0x%012lX   __alignof__() = %-3d   my_alignof() = %3d\n", (long) data4, (int) __alignof__( data4 ), my_alignof( data4 ));
  data5 = malloc_aligned( 25000, 128 );
  if( !data5 )
    error_exit( "An error occured while allocating Data5." );
  printf( "Data5:       0x%012lX   __alignof__() = %-3d   my_alignof() = %3d\n", (long) data5, (int) __alignof__( data5 ), my_alignof( data5 ));
  
  //  I need a much more extensive testing routine with an array of pointers to randomly aligned and sized memory chunks
  //  1000 pointers initialized to null, randomly allocated and freed
  printf( "-----------------------------------------------------------\n" );

  free_aligned( data4 );
  free_aligned( data2 );
  free_aligned( data1 );
  free_aligned( data3 );
  free_aligned( data5 );
/*
  free_aligned( data1 );
  free_aligned( data2 );
  free_aligned( data3 );
  free_aligned( data4 );
*/

  srand( time( NULL ));
  for( index = 0; index < ELEMENTS; index++ )
  {
    array[ index ] = NULL;
  }
  forever
  {
    index = rand() % ELEMENTS;
    if( array[ index ] )
    {
//      printf( "Array[%03d]:  0x%012lX   __alignof__() = %-4d   my_req_alignof() = %-4d   my_alignof() = %-4d  (count = %7d) (F)\n", 
//               index, (long) array[ index ], (int) __alignof__( array[ index ] ), my_req_alignof( array[ index ] ), my_alignof( array[ index ] ), count + 1 );
//      printf( "\n" );
      free_aligned( array[ index ] );
      array[ index ] = NULL;
    }
    else
    {
      int size;
      int alignment;

      size = rand() % 11;
      alignment = 1 << size;
      size = rand() % MAX_SIZE;
      array[ index ] = malloc_aligned( size, alignment );
      if( !array[ index ] )
        error_exit( "An error occured while allocating an array element." );
//      printf( "Array[%03d]:  0x%012lX   __alignof__() = %-4d   my_req_alignof() = %-4d   my_alignof() = %-4d  (count = %7d) (A)\n", 
//               index, (long) array[ index ], (int) __alignof__( array[ index ] ), my_req_alignof( array[ index ] ), my_alignof( array[ index ] ), count + 1 );
//      printf( "\n" );
      if( my_req_alignof( array[ index ] ) > my_alignof( array[ index ] ))
        error_exit( "Requested alignment is greater then actual alignment" );
    }
    if( ++count >= COUNT )
      break;
  }
  for( index = 0; index < ELEMENTS; index++ )
    if( array[ index ] != NULL )
      free_aligned( array[ index ] );
  clean_aligned();
//  printf( "%d  %d  %d  %d\n", my_alignof( 1 ), my_alignof( 2 ), my_alignof( 4 ), my_alignof( 48 ));
  return( 0 );
} 
  


#endif





