/*
 * Written and Copyright (C) 2006 by Luca Barbato
 * 
 * Altivec optimization for 15bit and 32bit pixel other depth pending
 * for questions or suggestions <lu_zero@gentoo.org>
 * 
 * Permission is hereby granted to Michael Jennings to license this code as
 * he sees fit. I'd prefer the GPL but he will choose the BSD. The debate
 * is moot as this is to become a part of the Eterm project, for which he is 
 * the primary author.  For users of this code I ask that any modifications
 * be released back into the community but with Michael Jennings chooses the
 * BSD license then that request has no backing in law.
 *
 * 
 * 
 */

#include "cmod.h"
#include <stdio.h>
#ifdef ETERM_ALTIVEC
#include <altivec.h>

#ifdef CONFIG_DARWIN
#define AVV(x...) (x)
#else
#define AVV(x...) {x}
#endif

#define DATA16 unsigned short

/* RGB 15 */
void
shade_ximage_15_altivec(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{

    union {
        unsigned short tmp_s[8] __attribute__ ((aligned(16)));
        unsigned char tmp_c[16] __attribute__ ((aligned(16)));
        vector signed short vec;
        } argb;
    int x, y, q;
    
    vector pixel *ptr = data; // assume data is aligned correctly.

    
    bpl %= w;

/*    	    
    //check for alignment, if the image is large enough for us to work on it
    if (data&15){
	    // call the c function for the initial data till it's aligned
    }
*/
    // splat the *m modifiers to a full vector
    
    if ((rm <= 255) && (gm <= 255) && (bm <= 255)) {
        /* No saturation */
	printf("no saturation\n");
        vector unsigned char mod;
        argb.tmp_c[0] = 1;    //alpha
	argb.tmp_c[1] = rm;   //red
	argb.tmp_c[2] = gm;   //green
	argb.tmp_c[3] = bm;   //blue
	
        //we use vec_splat
	mod = vec_splat((vector unsigned int)argb.vec, 0);
        for (y = h; --y >= 0;) {
            for (x = w/8; x > 0; x--) {
                vector unsigned char a, b;
                vector unsigned short ev, od;
                //unpack to byte
		
                a = vec_unpackh(*ptr);
		
                //multiply even
                ev = vec_sr(vec_mule(a,mod),vec_splat_u16(5));
		
                //multiply odd
                od = vec_sr(vec_mulo(a,mod),vec_splat_u16(5));

                //pack
                a = vec_pack(vec_mergeh(ev,od),vec_mergel(ev,od));
		                
                //unpack to byte
                b = vec_unpackl(*ptr);

                //multiply even
                ev = vec_sr(vec_mule(b,mod),vec_splat_u16(5));

                //multiply odd
                od = vec_sr(vec_mulo(b,mod),vec_splat_u16(5));

                //pack+saturate
                b = vec_pack(vec_mergeh(ev,od),vec_mergel(ev,od));

                //pack to pixel
                *ptr = vec_packpx((vector unsigned int)a,(vector unsigned int)b);

		ptr++;
            }
            
            for (x = w%8; --x >= 0;) {
                int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (b & 0x7c00) * rm;
                g = (b & 0x3e0) * gm;
                b = (b & 0x1f) * bm;
                ((DATA16 *) ptr)[x] = ((r >> 8) & 0x7c00)
                    | ((g >> 8) & 0x3e0)
                    | ((b >> 8) & 0x1f);
            }
            
            ptr += bpl; //what remains
        }
        
    } else {
        vector unsigned short mod;
	
        argb.tmp_s[0] = 1;    //alpha
	argb.tmp_s[1] = rm;   //red
	argb.tmp_s[2] = gm;   //green
	argb.tmp_s[3] = bm;   //blue

        // permute, we cannot splat
        mod = vec_perm((vector unsigned char)argb.vec,
	    (vector unsigned char)AVV(0,0,0,0),
	    (vector unsigned char)AVV(0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
				      0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07));
	printf("\n");
        for (y = h; --y >= 0;) {
	    int t;
            for (x = w/8,t = 0; x > 0; x--,t+=8) {
                vector unsigned char a, b;
                vector unsigned short hi, lo;
		vector unsigned int ev, od;
                
		//unpack to byte
                a = vec_unpackh(*ptr);
                
		//unpack to short
                hi = (vector unsigned short)vec_unpackh((vector signed char)a);

		//multiply and shift to right
		ev = vec_sr(vec_mule(hi,mod),vec_splat_u32(5));
		od = vec_sr(vec_mulo(hi,mod),vec_splat_u32(5));

		//pack to short
		hi = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));
		
		//unpack to short
                lo = (vector unsigned short)vec_unpackl((vector signed char)a);

                //multiply and shift to right
		ev = vec_sr(vec_mule(lo,mod),vec_splat_u32(5));
		od = vec_sr(vec_mulo(lo,mod),vec_splat_u32(5));
		
		//pack to short
		lo = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));

                //pack to byte
                a = vec_packsu(hi,lo);
                
                //unpack to byte 
                b = vec_unpackl(*ptr);

                //unpack to short
                hi = (vector unsigned short)vec_unpackh((vector signed char)b);

                //multiply
		ev = vec_sr(vec_mule(hi,mod),vec_splat_u32(5));
		od = vec_sr(vec_mulo(hi,mod),vec_splat_u32(5));
		
		hi = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));

                //unpack to short
                lo = (vector unsigned short)vec_unpackl((vector signed char)b);

                //multiply
		ev = vec_sr(vec_mule(lo,mod),vec_splat_u32(5));
		od = vec_sr(vec_mulo(lo,mod),vec_splat_u32(5));
		    
		lo = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));

                //pack to byte
                b = vec_packsu(hi,lo);

                //pack to pixel
                *ptr = vec_packpx(
			    (vector unsigned int)a,
			    (vector unsigned int)b);
                ptr++;
            }
            
            for (x = w%8; x > 0; x--) {
                int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (((b >> 10 ) & 0x001f )  * rm ) >> 8;
                r = ( r > 0x001f ) ? 0x7c00 : ( r << 10 );
                g = (((b >>  5 ) & 0x001f ) * gm ) >> 8;
                g = ( g > 0x001f ) ? 0x03e0 : ( g << 5 );
                b = (( b         & 0x001f ) * bm ) >> 8;
                b = ( b > 0x001f ) ? 0x001f : b;
                ((DATA16 *) ptr)[x] = (r|g|b);
            }
            
            ptr += bpl;
        }
    }
}


#if 0
/* RGB 16 */
void
shade_ximage_16_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    //
    union {
        unsigned short tmp_s[8] __attribute__ ((aligned(16)));
        unsigned char tmp_c[16] __attribute__ ((aligned(16)));
        vector signed short vec;
        } argb;

    vector unsigned char *ptr;
    vector unsigned char rotation;
    vector unsigned char mask_r;
    
    int x, y;

    ptr = (unsigned char *) data + (w * sizeof(DATA16));
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -w; x < 0; x++) {
            // A|    R|    G|    B
            // A|11111|00000|11111
            // R|RRRRG|GGGGG|BBBBB
            // B|RRRRR|GGGGG|GBBBB rot
            // 

            
            //rotate
            //select
            //rotate
            //select
            //select
            //mule&mulo
            //packsu
            //rot&sel


            //OR

            //pixel unpak
            //rotate&select 
            //as 15
            //
         /*       int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (b & 0xf800) * rm;
                g = (b & 0x7e0) * gm;
                b = (b & 0x1f) * bm;
                ((DATA16 *) ptr)[x] = ((r >> 8) & 0xf800)
                    | ((g >> 8) & 0x7e0)
                    | ((b >> 8) & 0x1f); */
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            for (x = -w; x < 0; x++) {

        /*
                int r, g, b;
                b = ((DATA16 *) ptr)[x];
                r = ( (b >> 11 )            * rm ) >> 8;
		r = ( r > 0x001f ) ? 0xf800 : ( r << 11 );
                g = (((b >>  5 ) & 0x003f ) * gm ) >> 8;
		g = ( g > 0x003f ) ? 0x07e0 : ( g << 5 );
                b = (( b         & 0x001f ) * bm ) >> 8;
		b = ( b > 0x001f ) ? 0x001f : b;
                ((DATA16 *) ptr)[x] = (r|g|b);*/
            }
            for (x=-(w%4), x > 0, x++) {

            }
            ptr += bpl;
        }
    }
}

#endif

/* RGB 32 */
void
shade_ximage_32_altivec(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    union {
        unsigned short tmp_s[8] __attribute__ ((aligned(16)));
        unsigned char tmp_c[16] __attribute__ ((aligned(16)));
        vector unsigned char vec;
        } argb;
    
    vector unsigned char *ptr;
    int x, y, q;

    bpl %= w;

    ptr = (vector unsigned char *) data;
    if ((rm <= 255) && (gm <= 255) && (bm <= 255)) {
        /* No saturation */
        vector unsigned char mod;
        argb.tmp_c[0] = 1;    //alpha
	argb.tmp_c[1] = rm;   //red
	argb.tmp_c[2] = gm;   //green
	argb.tmp_c[3] = bm;   //blue
	printf (" no saturation\n");
        //we use vec_splat
	mod = vec_splat((vector unsigned int)argb.vec, 0);

	for (y = h; --y >= 0;) {
            for (x = w/4; x > 0; x--) { // 4 pixel at time

                vector unsigned short ev, od;
		
                //multiply even
                ev = vec_sr(vec_mule(*ptr,mod),vec_splat_u16(8));
//		printf("ev  ");
//		for (q=0;q<8;q++) printf(" %d",((unsigned short *)&ev)[q]);
//		printf("\n");
                //multiply odd
                od = vec_sr(vec_mulo(*ptr,mod),vec_splat_u16(8));
//		printf("od  ");
//              for (q=0;q<8;q++) printf(" %d",((unsigned short *)&od)[q]);
//		printf("\n");
                //pack
                *ptr = vec_pack(vec_mergeh(ev,od),vec_mergel(ev,od));
//		printf("ptr ");
//		for (q=0;q<14;q++) printf(" %d",((unsigned char *)ptr)[q]);
//		printf("\n");
                ptr++;
            }
            for (x = (w%4-1)*4; x >=0; x -= 4) {
		unsigned char *ptr_c = (unsigned char *)ptr;
		
		ptr_c[x + 1] =
			(unsigned char) ((ptr_c[x + 1] * rm) >> 8);
                ptr_c[x + 2] =
			(unsigned char) ((ptr_c[x + 2] * gm) >> 8);
		ptr_c[x + 3] = 
			(unsigned char) ((ptr_c[x + 3] * bm) >> 8);
            }
            
            ptr += bpl;
        }
    } else {
        vector unsigned short mod;
	
        argb.tmp_s[0] = 1;    //alpha
	argb.tmp_s[1] = rm;   //red
	argb.tmp_s[2] = gm;   //green
	argb.tmp_s[3] = bm;   //blue

        // permute, we cannot splat
        mod = vec_perm((vector unsigned char)argb.vec,
	    (vector unsigned char)AVV(0,0,0,0),
	    (vector unsigned char)AVV(0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
				      0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07));
	
        for (y = h; --y >= 0;) {
            for (x = w/4; x > 0; x--) { // 4 pixel at time
                vector unsigned short hi, lo;
		vector unsigned int ev, od;

                //unpack to short
                hi = (vector unsigned short)
				vec_mergeh((vector unsigned char)AVV(0),
					    (vector unsigned char)*ptr);

		//multiply and shift to right
		ev = vec_sr(vec_mule(hi,mod),vec_splat_u32(8));
		od = vec_sr(vec_mulo(hi,mod),vec_splat_u32(8));
	
		//pack to short
		hi = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));

                //unpack to short
                lo = (vector unsigned short)
				vec_mergel((vector unsigned char)AVV(0),
					    (vector unsigned char)*ptr);

                //multiply and shift to right
		ev = vec_sr(vec_mule(lo,mod),vec_splat_u32(8));
		od = vec_sr(vec_mulo(lo,mod),vec_splat_u32(8));

		//pack to short
		lo = vec_packsu(vec_mergeh(ev,od),vec_mergel(ev,od));

                //pack to byte
                *ptr = vec_packsu(hi,lo);
		
		ptr++;
            }
            
            for (x = (w%4-1) * 4; x >= 0; x -= 4) {
                int r, g, b;
		unsigned char *ptr_c = (unsigned char *) ptr;
                r = (ptr_c[x + 1] * rm) >> 8;
                ptr_c[x + 1] = r|(!(r >> 8) - 1);
                g = (ptr_c[x + 2] * gm) >> 8;
                ptr_c[x + 2] = g|(!(g >> 8) - 1);
                b = (ptr_c[x + 3] * bm) >> 8;
                ptr_c[x + 3] = b|(!(b >> 8) - 1);
            }
            ptr += bpl;
        }
    }
}
#if 0
/* RGB 24 */
void
shade_ximage_24_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    unsigned char *ptr;
    int x, y;

    ptr = (unsigned char *) data + (w * 3);
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -(w * 3); x < 0; x += 3) {

# if WORDS_BIGENDIAN
                ptr[x + 0] = (ptr[x + 0] * rm) >> 8;
                ptr[x + 1] = (ptr[x + 1] * gm) >> 8;
                ptr[x + 2] = (ptr[x + 2] * bm) >> 8;
# else
                ptr[x + 2] = (ptr[x + 2] * rm) >> 8;
                ptr[x + 1] = (ptr[x + 1] * gm) >> 8;
                ptr[x + 0] = (ptr[x + 0] * bm) >> 8;
# endif
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            for (x = -(w * 3); x < 0; x += 3) {
                int r, g, b;
# if WORDS_BIGENDIAN
                r = (ptr[x + 0] * rm) >> 8;
                ptr[x + 0] = r|(!(r >> 8) - 1);
                g = (ptr[x + 1] * gm) >> 8;
                ptr[x + 1] = g|(!(g >> 8) - 1);
                b = (ptr[x + 2] * bm) >> 8;
                ptr[x + 2] = b|(!(b >> 8) - 1);
# else
                r = (ptr[x + 2] * rm) >> 8;
                ptr[x + 2] = r|(!(r >> 8) - 1);
                g = (ptr[x + 1] * gm) >> 8;
                ptr[x + 1] = g|(!(g >> 8) - 1);
                b = (ptr[x + 0] * bm) >> 8;
                ptr[x + 0] = b|(!(b >> 8) - 1);
# endif
            }
            ptr += bpl;
        }
    }
}
#endif
#endif
