
#include "cmod.h"

#define DATA16 unsigned short

/* RGB 15 */
void
shade_ximage_15_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    unsigned char *ptr;
    int x, y;

    ptr = (unsigned char *) data + (w * sizeof(DATA16));
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -w; x < 0; x++) {
                int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (b & 0x7c00) * rm;
                g = (b & 0x3e0) * gm;
                b = (b & 0x1f) * bm;
                ((DATA16 *) ptr)[x] = ((r >> 8) & 0x7c00)
                    | ((g >> 8) & 0x3e0)
                    | ((b >> 8) & 0x1f);
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            for (x = -w; x < 0; x++) {
                int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (((b >> 10 ) & 0x001f )  * rm ) >> 8;
                r = ( r > 0x001f ) ? 0x7c00 : ( r << 10 );
                g = (((b >>  5 ) & 0x001f ) * gm ) >> 8;
                g = ( g > 0x001f ) ? 0x03e0 : ( g << 5 );
                b = (( b         & 0x001f ) * bm ) >> 8;
                b = ( b > 0x001f ) ? 0x001f : b;
                ((DATA16 *) ptr)[x] = (r|g|b);
            }
            ptr += bpl;
        }
    }
}

/* RGB 16 */
void
shade_ximage_16_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    unsigned char *ptr;
    int x, y;

    ptr = (unsigned char *) data + (w * sizeof(DATA16));
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -w; x < 0; x++) {
                int r, g, b;

                b = ((DATA16 *) ptr)[x];
                r = (b & 0xf800) * rm;
                g = (b & 0x7e0) * gm;
                b = (b & 0x1f) * bm;
                ((DATA16 *) ptr)[x] = ((r >> 8) & 0xf800)
                    | ((g >> 8) & 0x7e0)
                    | ((b >> 8) & 0x1f);
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            int r, g, b;
            for (x = -w; x < 0; x++) {
                b = ((DATA16 *) ptr)[x];
                r = ( (b >> 11 )            * rm ) >> 8;
		r = ( r > 0x001f ) ? 0xf800 : ( r << 11 );
                g = (((b >>  5 ) & 0x003f ) * gm ) >> 8;
		g = ( g > 0x003f ) ? 0x07e0 : ( g << 5 );
                b = (( b         & 0x001f ) * bm ) >> 8;
		b = ( b > 0x001f ) ? 0x001f : b;
                ((DATA16 *) ptr)[x] = (r|g|b);
            }
            ptr += bpl;
        }
    }
}

/* RGB 32 */
void
shade_ximage_32_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    unsigned char *ptr;
    int x, y;

    ptr = (unsigned char *) data + (w * 4);
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -(w * 4); x < 0; x += 4) {
# if WORDS_BIGENDIAN
//		if (x%16) printf("\n scalar ");
                ptr[x + 1] = (unsigned char) ((ptr[x + 1] * rm) >> 8);
//		printf(" %d", ptr[x + 1]);
                ptr[x + 2] = (unsigned char) ((ptr[x + 2] * gm) >> 8);
//		printf(" %d", ptr[x + 2]);		
                ptr[x + 3] = (unsigned char) ((ptr[x + 3] * bm) >> 8);
//		printf(" %d", ptr[x + 3]);
//		if (x%16) printf("\n");
# else
                ptr[x + 2] = (unsigned char) ((ptr[x + 2] * rm) >> 8);
                ptr[x + 1] = (unsigned char) ((ptr[x + 1] * gm) >> 8);
                ptr[x + 0] = (unsigned char) ((ptr[x + 0] * bm) >> 8);
# endif
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            for (x = -(w * 4); x < 0; x += 4) {
                int r, g, b;
# if WORDS_BIGENDIAN
                r = (ptr[x + 1] * rm) >> 8;
                ptr[x + 1] = r|(!(r >> 8) - 1);
                g = (ptr[x + 2] * gm) >> 8;
                ptr[x + 2] = g|(!(g >> 8) - 1);
                b = (ptr[x + 3] * bm) >> 8;
                ptr[x + 3] = b|(!(b >> 8) - 1);
# else
                r = (ptr[x + 2] * rm) >> 8;
                ptr[x + 2] = r|(!(r >> 8) - 1);
                g = (ptr[x + 1] * gm) >> 8;
                ptr[x + 1] = g|(!(g >> 8) - 1);
                b = (ptr[x + 0] * bm) >> 8;
                ptr[x + 0] = b|(!(b >> 8) - 1);
# endif
            }
            ptr += bpl;
        }
    }
}

/* RGB 24 */
void
shade_ximage_24_C(void *data, int bpl, int w, int h, int rm, int gm, int bm)
{
    unsigned char *ptr;
    int x, y;

    ptr = (unsigned char *) data + (w * 3);
    if ((rm <= 256) && (gm <= 256) && (bm <= 256)) {
        /* No saturation */
        for (y = h; --y >= 0;) {
            for (x = -(w * 3); x < 0; x += 3) {

# if WORDS_BIGENDIAN
                ptr[x + 0] = (ptr[x + 0] * rm) >> 8;
                ptr[x + 1] = (ptr[x + 1] * gm) >> 8;
                ptr[x + 2] = (ptr[x + 2] * bm) >> 8;
# else
                ptr[x + 2] = (ptr[x + 2] * rm) >> 8;
                ptr[x + 1] = (ptr[x + 1] * gm) >> 8;
                ptr[x + 0] = (ptr[x + 0] * bm) >> 8;
# endif
            }
            ptr += bpl;
        }
    } else {
        for (y = h; --y >= 0;) {
            for (x = -(w * 3); x < 0; x += 3) {
                int r, g, b;
# if WORDS_BIGENDIAN
                r = (ptr[x + 0] * rm) >> 8;
                ptr[x + 0] = r|(!(r >> 8) - 1);
                g = (ptr[x + 1] * gm) >> 8;
                ptr[x + 1] = g|(!(g >> 8) - 1);
                b = (ptr[x + 2] * bm) >> 8;
                ptr[x + 2] = b|(!(b >> 8) - 1);
# else
                r = (ptr[x + 2] * rm) >> 8;
                ptr[x + 2] = r|(!(r >> 8) - 1);
                g = (ptr[x + 1] * gm) >> 8;
                ptr[x + 1] = g|(!(g >> 8) - 1);
                b = (ptr[x + 0] * bm) >> 8;
                ptr[x + 0] = b|(!(b >> 8) - 1);
# endif
            }
            ptr += bpl;
        }
    }
}

