/*  File: cmod.h
 *  Written and Copyright by Tres Melton tres <at> mindspring <dot> com
 *
 *
 *  Intel Chips		SIMD Instruction Set
 *  Pentium			None
 *  Pentium Pro			None
 *  Pentium MMX			MMX
 *  Pentium II			MMX
 *  Pentium III			SSE
 *  Pentium IV			SSE2
 *  Pentium IV (Prescott)	SSE3
 *  Pentium IV Xeon (EM64T)	SSE3
 *
 *  AMD Chips		SIMD Instruction Set
 *  AMD K6			MMX
 *  AMD K6-2			3D Now
 *  AMD K6-III			Extended 3D Now
 *  Athalon			3D Now
 *  Athalon XP			SSE
 *  AMD 64			SSE2
 *  AMD 64 (rev. E or Venice)	SSE3
 *  AMD 64 X2			SSE3
 *
 *  AMD SIMD Instruction Sets Includes
 *  3D Now			MMX
 *  Advanced/Extended 3D Now	MMX
 *  3D Now Pro			SSE
 *
 *
 *  Data Alignement (required for optimal performance):
 *  MMX ( 64 bits /  8 bytes)	gcc		__attribute__ ((aligned (8)))
 *				MS Visual C++	__declspec (align (8))
 *  Data Alignement (currently required to not segfalt):
 *  SSE (128 bits / 16 bytes)	gcc		__attribute__ ((aligned (16)))
 *				MS Visual C++	__declspec (align (16))
 *  Altivec ( 128 bits / 16 bytes) gcc 		__attribute__ ((aligned (16)))
 *  
 *  The MSVC Processor Pack Upgrade is required for memory alignement in MS 
 *  and the new call is _mm_malloc(sz,alignment) instead of malloc().  
 *  Remember to use _mm_free(); to free the code.  Reference:
 *  http://cfxweb.net/modules.php?name=News&file=article&sid=1316
 *
 *  The challenge is now on for the PowerPC gurus to adapt this code to use
 *  the PowerPC's Altivec SIMD engine to achieve the same performance 
 *  increases.  PA-RISC has the MAX extensions, the Alpha has MVI, the SGI
 *  has MDMX, and the UltraSPARC has VIS.  Good luck guys!  :-)
 *
 *  P.S.	Sorry MEJ.  :-/
 *
 *
 *  Note to self (prior to starting the SIMD instructions):
 *  Serious performance improvements can be had with the following:
 *    prefetchnta (%%esi)
 *    prefetcht0 0x400(%%esi)
 *    prefetcht1 0x10000(%%esi)
 *    prefetcht2 0x20000(%%esi)
 *  try (and use fence):
 *    prefetchnta (%%esi)
 *    prefetcht0 bpl(%%esi)
 *  And then add profiling to tst.c
 *
 *  Double check instruction availability on the various chips.  (SSE?)
 *  Pentium IV and AMD64 have them for sure.
 *
 *  Change the movdqu to movdqa  (unaligned to aligned)
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

#ifndef CMOD_H
#define CMOD_H 1

#define ETERM_ARCH_UNKNOWN	1
#define ETERM_ARCH_x86		201
#define ETERM_ARCH_x86_64	301

#define ETERM_SIMD_UNKNOWN	1
#define ETERM_SIMD_MMX		201
#define ETERM_SIMD_MMX_PLUS	202
#define ETERM_SIMD_SSE		301
#define ETERM_SIMD_SSE2 	302
#define ETERM_SIMD_SSE3		303
#define ETERM_ALTIVEC           501


//#define ETERM_ARCH      ETERM_ARCH_x86_64
//#define ETERM_SIMD      ETERM_SIMD_SSE2
//#define ETERM_ALIGNMENT 16
#define SIMD_USE_ALIGNED	0

//  The C routines.  Eventually these will go into the macros but for testing I need
//    both C and an SIMD engine to compare. 
extern void shade_ximage_15_C(void *data, int bpl, int w, int h, int rm, int gm, int bm);
extern void shade_ximage_16_C(void *data, int bpl, int w, int h, int rm, int gm, int bm);
extern void shade_ximage_24_C(void *data, int bpl, int w, int h, int rm, int gm, int bm);
extern void shade_ximage_32_C(void *data, int bpl, int w, int h, int rm, int gm, int bm);

#if     ETERM_ARCH == ETERM_ARCH_x86
#  if   ETERM_SIMD == ETERM_SIMD_MMX
     //  printf( "x86     MMX Routines.\n" );
     extern void shade_ximage_15_x86_mmx(void *data, int bpl, int w, int h, int rm, int gm, int bm);
     extern void shade_ximage_16_x86_mmx(void *data, int bpl, int w, int h, int rm, int gm, int bm);
     extern void shade_ximage_32_x86_mmx(void *data, int bpl, int w, int h, int rm, int gm, int bm);
#    define shade_ximage_15_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_15_x86_mmx((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_16_x86_mmx((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_32_x86_mmx((data), (bpl), (w), (h), (rm), (gm), (bm))
#  elif ETERM_SIMD == ETERM_SIMD_SSE
#    error "SSE on x86 has not been implemented yet.  Use the MMX code base"
     //  printf( "x86     SSE Routines.\n" );
     extern void shade_ximage_15_x86_sse_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_15_x86_sse_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_sse_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_sse_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_32_x86_sse  ( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
#    if SIMD_USE_ALIGNED
       //  For some reason the alignment stuff isn't working on x86/SSE2  :|
#      define shade_ximage_15_x86_sse( data, bpl, w, h, rm, gm, bm )                 \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_15_x86_sse_U((data), (bpl), (w), (h), (rm), (gm), (bm))      : \
         shade_ximage_15_x86_sse_A((data), (bpl), (w), (h), (rm), (gm), (bm));       \
       }
#      define shade_ximage_16_x86_sse( data, bpl, w, h, rm, gm, bm )                 \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_16_x86_sse_U((data), (bpl), (w), (h), (rm), (gm), (bm))      : \
         shade_ximage_16_x86_sse_A((data), (bpl), (w), (h), (rm), (gm), (bm));       \
       }
#    else  //  Use unaligned functions
#      define shade_ximage_15_x86_sse( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_15_x86_sse_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#      define shade_ximage_16_x86_sse( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_16_x86_sse_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#    endif
#    define shade_ximage_15_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_15_x86_sse((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_16_x86_sse((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32_SIMD( data, bpl, w, h, rm, gm, bm )         \
       shade_ximage_32_x86_sse((data), (bpl), (w), (h), (rm), (gm), (bm))
#  elif ETERM_SIMD == ETERM_SIMD_SSE2
     //  printf( "x86    SSE2 Routines.\n" );
     extern void shade_ximage_15_x86_sse2_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_15_x86_sse2_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_sse2_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_sse2_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_32_x86_sse2  ( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
#    if SIMD_USE_ALIGNED
       //  For some reason the alignment stuff isn't working on x86/SSE2  :|
#      define shade_ximage_15_x86_sse2( data, bpl, w, h, rm, gm, bm )                \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_15_x86_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))     : \
         shade_ximage_15_x86_sse2_A((data), (bpl), (w), (h), (rm), (gm), (bm));      \
       }
#      define shade_ximage_16_x86_sse2( data, bpl, w, h, rm, gm, bm )                \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_16_x86_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))     : \
         shade_ximage_16_x86_sse2_A((data), (bpl), (w), (h), (rm), (gm), (bm));      \
       }
#    else  //  Use unaligned functions
#      define shade_ximage_15_x86_sse2( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_15_x86_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#      define shade_ximage_16_x86_sse2( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_16_x86_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#    endif
#    define shade_ximage_15_SIMD( data, bpl, w, h, rm, gm, bm )          \
       shade_ximage_15_x86_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16_SIMD( data, bpl, w, h, rm, gm, bm )          \
       shade_ximage_16_x86_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32_SIMD( data, bpl, w, h, rm, gm, bm )          \
       shade_ximage_32_x86_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#  else  //  Unknown SIMD for x86, C Routines.
#    define shade_ximage_15( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_15_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_16_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_32_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#  endif
#elif   ETERM_ARCH == ETERM_ARCH_x86_64
#  if   ETERM_SIMD == ETERM_SIMD_SSE2
     //  printf( "x86-64 SSE2 Routines.\n" );
     //  Eventually ETERM_ALIGNMENT will be set in the macros here for Eterm.  
     //    For testing though it needs to be easily adjustable and is at the top.
     //#define ETERM_ALIGNMENT 16
     extern void shade_ximage_15_x86_64_sse2_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_15_x86_64_sse2_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_64_sse2_U( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_16_x86_64_sse2_A( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
     extern void shade_ximage_32_x86_64_sse2  ( volatile void *data, volatile int bpl, volatile int w, volatile int h, volatile int rm, volatile int gm, volatile int bm );
#    if SIMD_USE_ALIGNED
#      define shade_ximage_15_x86_64_sse2( data, bpl, w, h, rm, gm, bm )             \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_15_x86_64_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))  : \
         shade_ximage_15_x86_64_sse2_A((data), (bpl), (w), (h), (rm), (gm), (bm));   \
       }
#      define shade_ximage_16_x86_64_sse2( data, bpl, w, h, rm, gm, bm )             \
       {                                                                             \
         (((long) ( data )) & ((long) ( bpl )) & ((long) ( ETERM_ALIGNMENT - 1 ))) ? \
         shade_ximage_16_x86_64_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))  : \
         shade_ximage_16_x86_64_sse2_A((data), (bpl), (w), (h), (rm), (gm), (bm));   \
       }
#    else  //  Use unaligned functions
#      define shade_ximage_15_x86_64_sse2( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_15_x86_64_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#      define shade_ximage_16_x86_64_sse2( data, bpl, w, h, rm, gm, bm )        \
         shade_ximage_16_x86_64_sse2_U((data), (bpl), (w), (h), (rm), (gm), (bm))
#    endif
#    define shade_ximage_15_SIMD( data, bpl, w, h, rm, gm, bm )             \
       shade_ximage_15_x86_64_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16_SIMD( data, bpl, w, h, rm, gm, bm )             \
       shade_ximage_16_x86_64_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32_SIMD( data, bpl, w, h, rm, gm, bm )             \
       shade_ximage_32_x86_64_sse2((data), (bpl), (w), (h), (rm), (gm), (bm))
#  else  //  Other, lesser, x86-64 combinations make no sense, C Routines.
#    define shade_ximage_15( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_15_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_16( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_16_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#    define shade_ximage_32( data, bpl, w, h, rm, gm, bm )        \
       shade_ximage_32_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#  endif
#else  //  Unknown Arch,  C Routines.
#  define shade_ximage_15( data, bpl, w, h, rm, gm, bm )        \
     shade_ximage_15_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#  define shade_ximage_16( data, bpl, w, h, rm, gm, bm )        \
     shade_ximage_16_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#  define shade_ximage_32( data, bpl, w, h, rm, gm, bm )        \
     shade_ximage_32_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#endif



#if 0  //  Eventually define a macro for all the base functions
#define shade_ximage_15( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_15_((data), (bpl), (w), (h), (rm), (gm), (bm))
#define shade_ximage_16( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_16_((data), (bpl), (w), (h), (rm), (gm), (bm))
#define shade_ximage_32( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_32_((data), (bpl), (w), (h), (rm), (gm), (bm))
#endif

#if ETERM_SIMD == ETERM_ALTIVEC  //  Altivec stuff
#define shade_ximage_15_SIMD( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_15_altivec((data), (bpl), (w), (h), (rm), (gm), (bm))
#define shade_ximage_16_SIMD( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_16_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#define shade_ximage_32_SIMD( data, bpl, w, h, rm, gm, bm )       \
   shade_ximage_32_altivec((data), (bpl), (w), (h), (rm), (gm), (bm))
#endif


//  24bpp is an unfortunate choice as there are no SIMD accelerators for it.  :-(
#define shade_ximage_24_SIMD( data, bpl, w, h, rm, gm, bm )   \
   shade_ximage_24_C((data), (bpl), (w), (h), (rm), (gm), (bm))
#endif	/*  CMOD_H  */

