/*
 * Copyright (C) 1997-2004, Michael Jennings
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies of the Software, its documentation and marketing & publicity
 * materials, and acknowledgment shall be given in the documentation, materials
 * and software packages that this Software was used.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* #include "config.h" */

/* MMX routines for tinting XImages written by Willem Monsuwe <willem@stack.nl> */

/* Function calling conventions:
 *   shade_ximage_xx(void *data, int bpl, int w, int h, int rm, int gm, int bm);
 */

/* #ifdef HAVE_MMX_32 */

#if ETERM_ARCH == ETERM_ARCH_x86 && ETERM_SIMD == ETERM_SIMD_SSE

#define data	8(%ebp)
#define bpl	12(%ebp)
#define w	16(%ebp)
#define h	20(%ebp)
#define rm	24(%ebp)
#define gm	28(%ebp)
#define bm	32(%ebp)

.global shade_ximage_15_x86_sse
        .type shade_ximage_15_x86_sse,@function
.global shade_ximage_16_x86_sse
        .type shade_ximage_16_x86_sse,@function
.global shade_ximage_32_x86_sse
        .type shade_ximage_32_x86_sse,@function

.bss
.text
.align 16

#define ENTER                   \
        pushl %ebp              ;\
        movl %esp, %ebp         ;\
        pushl %ebx              ;\
        pushl %ecx              ;\
        pushl %edx              ;\
        pushl %edi              ;\
        pushl %esi              ;\
        movl data, %esi         ;\
        movl w, %ebx            ;\
        movl h, %edx

#define LEAVE                   \
4:                              ;\
        emms                    ;\
        popl %esi               ;\
        popl %edi               ;\
        popl %edx               ;\
        popl %ecx               ;\
        popl %ebx               ;\
        movl %ebp, %esp         ;\
        popl %ebp               ;\
        ret


shade_ximage_15_x86_sse:
        ENTER

        leal -14(%esi, %ebx, 2), %esi
        negl %ebx
        jz 5f

        /* Setup multipliers */
        movd rm, %xmm5
        movd gm, %xmm6
        movd bm, %xmm7
        shufps $0x0, %xmm5, %xmm5
        shufps $0x0, %xmm6, %xmm6
        shufps $0x0, %xmm7, %xmm7
        punpcklwd %xmm5, %xmm5    /* 00 00 00 00 rm rm rm rm */
        punpcklwd %xmm6, %xmm6    /* 00 00 00 00 gm gm gm gm */
        punpcklwd %xmm7, %xmm7    /* 00 00 00 00 bm bm bm bm */
///        punpckldq %xmm5, %xmm5    /* rm rm rm rm rm rm rm rm */
///        punpckldq %xmm6, %xmm6    /* gm gm gm gm gm gm gm gm */
///        punpckldq %xmm7, %xmm7    /* bm bm bm bm bm bm bm bm */
//        punpcklqdq %xmm5, %xmm5    /* rm rm rm rm rm rm rm rm rm rm rm rm rm rm rm rm */
//        punpcklqdq %xmm6, %xmm6    /* gm gm gm gm gm gm gm gm gm gm gm gm gm gm gm gm */
//        punpcklqdq %xmm7, %xmm7    /* bm bm bm bm bm bm bm bm bm bm bm bm bm bm bm bm */
//        shufps $0x0, %xmm5, %xmm5
//        shufps $0x0, %xmm6, %xmm6
//        shufps $0x0, %xmm7, %xmm7
//	movlhps

        cmpl $256, rm
        jg shade_ximage_15_x86_sse_saturate
        cmpl $256, gm
        jg shade_ximage_15_x86_sse_saturate
        cmpl $256, bm
        jg shade_ximage_15_x86_sse_saturate

1:      movl %ebx, %ecx
        addl $7, %ecx
        jns 3f
2:
        movdqu (%esi, %ecx, 2), %xmm0

        movdqu %xmm0, %xmm1         /* rg gb */
        movdqu %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $10, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $11, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $3, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* 00 0r */
        pmulhw %xmm6, %xmm1       /* 00 0g */
        pmulhw %xmm7, %xmm2       /* 00 0b */

        psllw $10, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */
        
        movdqu %xmm0, (%esi, %ecx, 2)

        addl $8, %ecx
        js 2b
        jmp 4f
3:
        movw (%esi, %ecx, 2), %ax
        movd %eax, %xmm0

        movq %xmm0, %xmm1         /* rg gb */
        movq %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $10, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $11, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $3, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* 00 0r */
        pmulhw %xmm6, %xmm1       /* 00 0g */
        pmulhw %xmm7, %xmm2       /* 00 0b */

        psllw $10, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movd %xmm0, %eax
        movw %ax, (%esi, %ecx, 2)

        incl %ecx
4:
        cmpl $6, %ecx
        jng 3b

        addl bpl, %esi
        decl %edx
        jnz 1b
5:
        LEAVE


shade_ximage_15_x86_sse_saturate:

        pcmpeqw %xmm3, %xmm3
        psllw $5, %xmm3          /* ff e0 ff e0 ff e0 ff e0 */

1:      movl %ebx, %ecx
        addl $7, %ecx
        jns 3f
2:
        movdqu (%esi, %ecx, 2), %xmm0

        movdqu %xmm0, %xmm1         /* rg gb */
        movdqu %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $10, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $11, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $3, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* xx xr */
        pmulhw %xmm6, %xmm1       /* xx xg */
        pmulhw %xmm7, %xmm2       /* xx xb */

        /* Saturate upper */
        paddusw %xmm3, %xmm0      /* ff er */
        paddusw %xmm3, %xmm1      /* ff eg */
        paddusw %xmm3, %xmm2      /* ff eb */

        psubw %xmm3, %xmm0        /* 00 0r */
        psubw %xmm3, %xmm1        /* 00 0g */
        psubw %xmm3, %xmm2        /* 00 0b */
        
        psllw $10, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movdqu %xmm0, (%esi, %ecx, 2)

        addl $8, %ecx
        js 2b
        jmp 4f
3:
        movw (%esi, %ecx, 2), %ax
        movd %eax, %xmm0

        movq %xmm0, %xmm1         /* rg gb */
        movq %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $10, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $11, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $3, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* xx xr */
        pmulhw %xmm6, %xmm1       /* xx xg */
        pmulhw %xmm7, %xmm2       /* xx xb */

        /* Saturate upper */
        paddusw %xmm3, %xmm0      /* ff er */
        paddusw %xmm3, %xmm1      /* ff eg */
        paddusw %xmm3, %xmm2      /* ff eb */

        psubw %xmm3, %xmm0        /* 00 0r */
        psubw %xmm3, %xmm1        /* 00 0g */
        psubw %xmm3, %xmm2        /* 00 0b */
        
        psllw $10, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movd %xmm0, %eax
        movw %ax, (%esi, %ecx, 2)

        incl %ecx
4:
        cmpl $6, %ecx
        jng 3b

        addl bpl, %esi
        decl %edx
        jnz 1b
5:
        LEAVE


shade_ximage_16_x86_sse:
        ENTER

        leal -14(%esi, %ebx, 2), %esi
        negl %ebx
        jz 5f

        /* Setup multipliers */
        movd rm, %xmm5
        movd gm, %xmm6
        movd bm, %xmm7
        punpcklwd %xmm5, %xmm5    /* 00 00 00 00 rm rm rm rm */
        punpcklwd %xmm6, %xmm6    /* 00 00 00 00 gm gm gm gm */
        punpcklwd %xmm7, %xmm7    /* 00 00 00 00 bm bm bm bm */
        punpckldq %xmm5, %xmm5    /* rm rm rm rm rm rm rm rm */
        punpckldq %xmm6, %xmm6    /* gm gm gm gm gm gm gm gm */
        punpckldq %xmm7, %xmm7    /* bm bm bm bm bm bm bm bm */
//        punpcklqdq %xmm5, %xmm5    /* rm rm rm rm rm rm rm rm rm rm rm rm rm rm rm rm */
//        punpcklqdq %xmm6, %xmm6    /* gm gm gm gm gm gm gm gm gm gm gm gm gm gm gm gm */
//        punpcklqdq %xmm7, %xmm7    /* bm bm bm bm bm bm bm bm bm bm bm bm bm bm bm bm */  /*  SSE2 only  */
//        shufps $0x0, %xmm5, %xmm5
//        shufps $0x0, %xmm6, %xmm6
//        shufps $0x0, %xmm7, %xmm7

        cmpl $256, rm
        jg shade_ximage_16_x86_sse_saturate
        cmpl $256, gm
        jg shade_ximage_16_x86_sse_saturate
        cmpl $256, bm
        jg shade_ximage_16_x86_sse_saturate

1:      movl %ebx, %ecx
        addl $7, %ecx
        jns 3f
2:
        movdqu (%esi, %ecx, 2), %xmm0

        movdqu %xmm0, %xmm1         /* rg gb */
        movdqu %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $11, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $10, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $2, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* 00 0r */
        pmulhw %xmm6, %xmm1       /* 00 0g */
        pmulhw %xmm7, %xmm2       /* 00 0b */

        psllw $11, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */
        
        movdqu %xmm0, (%esi, %ecx, 2)

        addl $8, %ecx
        js 2b
	jmp 4f
3:
        movw (%esi, %ecx, 2), %ax
        movd %eax, %xmm0

        movq %xmm0, %xmm1         /* rg gb */
        movq %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $11, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $10, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $2, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* 00 0r */
        pmulhw %xmm6, %xmm1       /* 00 0g */
        pmulhw %xmm7, %xmm2       /* 00 0b */

        psllw $11, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movd %xmm0, %eax
        movw %ax, (%esi, %ecx, 2)

        incl %ecx
4:
        cmpl $6, %ecx
        jng 3b

        addl bpl, %esi
        decl %edx
        jnz 1b
5:
        LEAVE


shade_ximage_16_x86_sse_saturate:

        pcmpeqw %xmm3, %xmm3
        movdqu %xmm3, %xmm4
        psllw $5, %xmm3          /* ff e0 ff e0 ff e0 ff e0 */
        psllw $6, %xmm4          /* ff c0 ff c0 ff c0 ff c0 */

1:      movl %ebx, %ecx
        addl $7, %ecx
        jns 3f
2:
        movdqu (%esi, %ecx, 2), %xmm0

        movdqu %xmm0, %xmm1         /* rg gb */
        movdqu %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $11, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $10, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $2, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* xx xr */
        pmulhw %xmm6, %xmm1       /* xx xg */
        pmulhw %xmm7, %xmm2       /* xx xb */

        /* Saturate upper */
        paddusw %xmm3, %xmm0      /* ff er */
        paddusw %xmm4, %xmm1      /* ff cg */
        paddusw %xmm3, %xmm2      /* ff eb */

        psubw %xmm4, %xmm1        /* 00 0g */
        psubw %xmm3, %xmm2        /* 00 0b */
        
        psllw $11, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movdqu %xmm0, (%esi, %ecx, 2)

        addl $8, %ecx
        js 2b
        jmp 4f
3:
        movw (%esi, %ecx, 2), %ax
        movd %eax, %xmm0

        movq %xmm0, %xmm1         /* rg gb */
        movq %xmm0, %xmm2         /* rg gb */
        psrlw $5, %xmm1          /* 0r rg */
        psrlw $11, %xmm0         /* 00 0r */
        psllw $11, %xmm2         /* b0 00 */
        psllw $10, %xmm1         /* g0 00 */
        psllw $8, %xmm0          /* 0r 00 */
        psrlw $2, %xmm1          /* 0g 00 */
        psrlw $3, %xmm2          /* 0b 00 */

        pmulhw %xmm5, %xmm0       /* xx xr */
        pmulhw %xmm6, %xmm1       /* xx xg */
        pmulhw %xmm7, %xmm2       /* xx xb */

        /* Saturate upper */
        paddusw %xmm3, %xmm0      /* ff er */
        paddusw %xmm4, %xmm1      /* ff cg */
        paddusw %xmm3, %xmm2      /* ff eb */

        psubw %xmm4, %xmm1        /* 00 0g */
        psubw %xmm3, %xmm2        /* 00 0b */
        
        psllw $11, %xmm0         /* r0 00 */
        psllw $5, %xmm1          /* 0g g0 */
        por %xmm2, %xmm0          /* r0 0b */
        por %xmm1, %xmm0          /* rg gb */

        movd %xmm0, %eax
        movw %ax, (%esi, %ecx, 2)

        incl %ecx
4:
        cmpl $6, %ecx
        jng 3b

        addl bpl, %esi
        decl %edx
        jnz 1b
5:
        LEAVE


shade_ximage_32_x86_sse:
        ENTER

        leal (%esi, %ebx, 4), %esi
        negl %ebx
        jz 3f

        movd rm, %xmm4
        movd gm, %xmm5
        movd bm, %xmm6
        psllq $32, %xmm4
        psllq $16, %xmm5
        por %xmm6, %xmm4
        por %xmm5, %xmm4

        pcmpeqw %xmm6, %xmm6
        psllw $15, %xmm6                 /* 80 00 80 00 80 00 80 00 */
        movq %xmm6, %xmm5
        pmulhw %xmm4, %xmm5               /* Get correction factor */
1:
        movl %ebx, %ecx
2:
        movd (%esi, %ecx, 4), %xmm1      /* 00 rr gg bb */
        pxor %xmm0, %xmm0
        punpcklbw %xmm1, %xmm0            /* 00 00 rr 00 gg 00 bb 00 */
        pxor %xmm6, %xmm0                 /* Flip sign */

        pmulhw %xmm4, %xmm0               /* 00 00 xx rr xx gg xx bb */
        psubw %xmm5, %xmm0                /* Correct range */
        packuswb %xmm0, %xmm0             /* 00 rr gg bb 00 rr gg bb */

        movd %xmm0, (%esi, %ecx, 4)

        incl %ecx
        jnz 2b

        addl bpl, %esi
        decl %edx
        jnz 1b
3:
        LEAVE

#endif
