//  File: test-it.c

//  Written & Copyright by Tres Melton 2005 - 2006

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include "time-it.h"

static unsigned long long start_cycle;
static unsigned long long cycle_count_overhead;
#ifdef ETERM_ARCH_PPC
static inline uint64_t access_counter(void)
{
    uint32_t tbu, tbl, temp;

     /* from section 2.2.1 of the 32-bit PowerPC PEM */
     __asm__ __volatile__(
         "1:\n"
         "mftbu  %2\n"
         "mftb   %0\n"
         "mftbu  %1\n"
         "cmpw   %2,%1\n"
         "bne    1b\n"
     : "=r"(tbl), "=r"(tbu), "=r"(temp)
     :
     : "cc");

     return (((uint64_t)tbu)<<32) | (uint64_t)tbl;
}
#elif ETERM_ARCH_x86 || ETERM_ARCH_x86_64
//  unsigned long long should be 64 bits on x86 & x86-64
static inline uint64_t access_counter( void )
{
  register unsigned eax_lo, edx_hi;

  __asm__(
	"subl	%%eax, %%eax		\n\t"		/* Zero eax -- is this faster than xor?  Input op for cpuid.  Cycles subtracted out in the end.  */
	"cpuid				\n\t"		/* Serialize the chip  */
	"rdtsc; 			\n\t"		/* Read cycle counter */
	"movl %%edx, %0		 	\n\t"		/*   and save the results */
	"movl %%eax, %1			\n\t"
	: "=r" (edx_hi), "=r" (eax_lo)			/* the hi and lo 32 bits of a 64 bit integer */
	: /* No input */
	: "%eax", "%ebx", "%ecx", "%edx"		/* the two outputs of rdtsc and the registers trashed by cpuid */
  );
  return((((unsigned long long) edx_hi ) << 32 ) | eax_lo );
}
#else
#error "Put your arch dependend code here"
#endif
//  The literature states that:  A strange quirk of the CPUID
//    instruction is that it can take longer to complete the first couple of times it is called. Thus, the best policy is to call the
//    instruction three times, measure the elapsed time on the third call, then subtract this measurement from all future measurements.
void access_counter_init( void )
{
  uint64_t temp1, temp2;
  
  access_counter();
  access_counter();
  temp1 = access_counter();
  temp2 = access_counter();
  cycle_count_overhead = temp2 - temp1;
}

void start_counter( void )
{
  start_cycle = access_counter();
}

long long get_counter( void )
{
  uint64_t end_cycle;
  
  end_cycle = access_counter();
  return(( end_cycle - start_cycle ) - cycle_count_overhead );
}

